@@ -1,10 +1,11 @@
 -- | Core types and definitions
 {-# LANGUAGE GeneralizedNewtypeDeriving, FlexibleContexts,
-             ExistentialQuantification, RankNTypes, DeriveDataTypeable,
+             ExistentialQuantification, RankNTypes, DeriveDataTypeable, NoMonomorphismRestriction,
              DeriveGeneric #-}
 module Test.Tasty.Core where
 
 import Control.Exception
+import Test.Tasty.Providers.ConsoleFormat
 import Test.Tasty.Options
 import Test.Tasty.Patterns
 import Test.Tasty.Patterns.Types
@@ -65,6 +66,17 @@ data Result = Result
     -- @FAIL@.
   , resultTime :: Time
     -- ^ How long it took to run the test, in seconds.
+  , resultDetailsPrinter :: ResultDetailsPrinter
+    -- ^ An action that prints additional information about a test.
+    --
+    -- This is similar to 'resultDescription' except it can produce
+    -- colorful/formatted output; see "Test.Tasty.Providers.ConsoleFormat".
+    --
+    -- This can be used instead of or in addition to 'resultDescription'.
+    --
+    -- Usually this is set to 'noResultDetails', which does nothing.
+    --
+    -- @since 1.3.1
   }
   deriving Show
 
@@ -104,6 +116,7 @@ exceptionResult e = Result
   , resultDescription = "Exception: " ++ show e
   , resultShortDescription = "FAIL"
   , resultTime = 0
+  , resultDetailsPrinter = noResultDetails
   }
 
 -- | Test progress information.
@@ -297,9 +310,9 @@ after deptype s =
 -- indroduced.
 data TreeFold b = TreeFold
   { foldSingle :: forall t . IsTest t => OptionSet -> TestName -> t -> b
-  , foldGroup :: TestName -> b -> b
-  , foldResource :: forall a . ResourceSpec a -> (IO a -> b) -> b
-  , foldAfter :: DependencyType -> Expr -> b -> b
+  , foldGroup :: OptionSet -> TestName -> b -> b
+  , foldResource :: forall a . OptionSet -> ResourceSpec a -> (IO a -> b) -> b
+  , foldAfter :: OptionSet -> DependencyType -> Expr -> b -> b
   }
 
 -- | 'trivialFold' can serve as the basis for custom folds. Just override
@@ -316,9 +329,9 @@ data TreeFold b = TreeFold
 trivialFold :: Monoid b => TreeFold b
 trivialFold = TreeFold
   { foldSingle = \_ _ _ -> mempty
-  , foldGroup = const id
-  , foldResource = \_ f -> f $ throwIO NotRunningTests
-  , foldAfter = \_ _ b -> b
+  , foldGroup = \_ _ b -> b
+  , foldResource = \_ _ f -> f $ throwIO NotRunningTests
+  , foldAfter = \_ _ _ b -> b
   }
 
 -- | Fold a test tree into a single value.
@@ -340,7 +353,7 @@ trivialFold = TreeFold
 -- affected by the subsequent option changes. This shouldn't be a problem
 -- in practice; OTOH, this behaviour may be changed later.
 foldTestTree
-  :: Monoid b
+  :: forall b . Monoid b
   => TreeFold b
      -- ^ the algebra (i.e. how to fold a tree)
   -> OptionSet
@@ -349,21 +362,23 @@ foldTestTree
      -- ^ the tree to fold
   -> b
 foldTestTree (TreeFold fTest fGroup fResource fAfter) opts0 tree0 =
-  let pat = lookupOption opts0
-  in go pat mempty opts0 tree0
+  go mempty opts0 tree0
   where
-    go pat path opts tree1 =
+    go :: (Seq.Seq TestName -> OptionSet -> TestTree -> b)
+    go path opts tree1 =
       case tree1 of
         SingleTest name test
           | testPatternMatches pat (path Seq.|> name)
             -> fTest opts name test
           | otherwise -> mempty
         TestGroup name trees ->
-          fGroup name $ foldMap (go pat (path Seq.|> name) opts) trees
-        PlusTestOptions f tree -> go pat path (f opts) tree
-        WithResource res0 tree -> fResource res0 $ \res -> go pat path opts (tree res)
-        AskOptions f -> go pat path opts (f opts)
-        After deptype dep tree -> fAfter deptype dep $ go pat path opts tree
+          fGroup opts name $ foldMap (go (path Seq.|> name) opts) trees
+        PlusTestOptions f tree -> go path (f opts) tree
+        WithResource res0 tree -> fResource opts res0 $ \res -> go path opts (tree res)
+        AskOptions f -> go path opts (f opts)
+        After deptype dep tree -> fAfter opts deptype dep $ go path opts tree
+      where
+        pat = lookupOption opts :: TestPattern
 
 -- | Get the list of options that are relevant for a given test tree
 treeOptions :: TestTree -> [OptionDescription]
